=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Aug 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__CurviShear.rb
# Original Date	:   23 Aug 2011 (based on work published in June 2010)
# Description	:   Shear curve(s) along their path (curvilenear shear) - Useful for ramps
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module CurviShear

#Texts for the module
T7[:DLG_CVS_PromptHeight] = "Top Height"
T7[:DLG_CVS_PromptBaseHeight] = "Base Height (constant)"
T7[:DLG_CVS_ReverseCurve] = "Reverse curve orientation"
T7[:DLG_CVS_PromptGroup] = "Generate in a Group"
T7[:DLG_CVS_NoCurve] = "NO curve in the selection"
T7[:DLG_CVS_Elevation] = "Highlighted edges indicate the side which will be elevated unless you change the orientation"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	execute
end

#----------------------------------------------------------------------------------------------------
# Texts and Messages
#----------------------------------------------------------------------------------------------------

def self.init_text
	@menutitle = T7[:PlugName]
	@prompt_height = T7[:DLG_CVS_PromptHeight]
	@prompt_base_height = T7[:DLG_CVS_PromptBaseHeight]
	@prompt_reverse = T7[:DLG_CVS_ReverseCurve]
	@prompt_group = T7[:DLG_CVS_PromptGroup]
	@dlg_yes = T7[:T_Yes]
	@dlg_no = T7[:T_No]
	@msg_no_curves = T7[:DLG_CVS_NoCurve]
	@msg_elevation = T7[:DLG_CVS_Elevation]
	
	@dlg_yesno = @dlg_yes + '|' + @dlg_no
end

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

#Check if the seelction contains curves and return their points
def self.contains_curves?(selection)
	hcurve = {}
	selection.each do |e|
		next if e.class != Sketchup::Edge
		curve = e.curve
		next unless curve
		hcurve[curve.entityID] = curve
	end
	return nil if hcurve.length == 0
	hcurve.values
end

#Initialize and execute the curve shearing
def self.execute(curves=nil)
	init_text
	@model = Sketchup.active_model
	selection = @model.selection
	unless curves
		curves = contains_curves?(selection)
		return UI.messagebox("#{@msg_no_curves}") unless curves
	end
	lpts = curves.collect { |curve| curve.vertices.collect { |v| v.position } }
	
	#Orient the curves
	ledges = []
	depart = lpts[0].first
	for i in 0..lpts.length-1
		pts = lpts[i]
		if pts[0].distance(depart) > pts[-1].distance(depart)
			lpts[i] = pts.reverse
			ipos = 0
		else
			ipos = -1
		end	
		ipos = (ipos+1).modulo(2) if @reverse
		ledges.push curves[i].edges[ipos]
	end
	
	#Handling the selection to highlight extremities of curves being moved
	selection.clear
	selection.add ledges
	Sketchup.set_status_text @msg_elevation
	
	#Get the vector and height
	h, bh = ask_parameters
	return unless h
	vec_dec = Z_AXIS
	lpts = lpts.collect { |pts| pts.reverse } if @reverse
	
	#Create the list of the sheared curves
	llpt = []
	lpts.each do |pts|
		llpt.push curvishear_curve(pts, vec_dec, h, bh)
	end
			
	#processing the curve and segment generation
	Sketchup.set_status_text T7[:MSG_Processing], SB_VCB_LABEL
	@model.start_operation @menutitle
	entities = @model.active_entities
	if @group
		g = entities.add_group
		entities = g.entities
	end
	
	llpt.each_with_index do |pts, i|
		entities.add_curve pts
		entities.add_edges [pts.first, lpts[i].first]
		entities.add_edges [pts.last, lpts[i].last]
	end
	for i in 1..llpt.length-1
		entities.add_edges [lpts[i-1].first, lpts[i].first]
		entities.add_edges [lpts[i-1].last, lpts[i].last]
		entities.add_edges [llpt[i-1].first, llpt[i].first]
		entities.add_edges [llpt[i-1].last, llpt[i].last]
	end	
	if @group
		lpts.each_with_index do |pts, i|
			entities.add_curve pts
		end
	end
	
	@model.commit_operation
	Sketchup.set_status_text T7[:MSG_Finished], SB_VCB_LABEL
	selection.clear
end
		
#Shear the curve with a fixed base height (bh) and a variable height (max = h)
def self.curvishear_curve(pts, vec, h, bh)
	dist = curvi_dist pts
	dtot = dist.last
	
	lptnew = []
	for i in 0.. pts.length-1
		lptnew.push pts[i].offset(vec, bh + h * dist[i] / dtot)
	end
	lptnew
end
	
#compute curvilinear distance for a sequence of points	
def self.curvi_dist(pts)
	dist = [0]
	d = 0
	for i in 1..pts.length-1
		d += pts[i].distance pts[i-1]
		dist[i] = d
	end
	dist
end

#dialog box for parameters	
def self.ask_parameters
	prompts = [@prompt_height, @prompt_base_height, @prompt_reverse, @prompt_group]
	unless @height
		@base_height = @height = ORIGIN.distance(ORIGIN) unless @height
		@group = false
		@reverse = false
	end	
	val_rev = (@reverse)? @dlg_yes : @dlg_no
	val_group = (@group)? @dlg_yes : @dlg_no
	results = [@height, @base_height, val_rev, val_group]
	while true
		results = UI.inputbox prompts, results, [nil, nil, @dlg_yesno, @dlg_yesno], @menutitle
		return nil unless results
		break
	end
	h, bh, ynr, yng = results
	return nil if h == 0 && bh == 0
	@height = h
	@base_height = bh
	@group = (yng == @dlg_yes)
	@reverse = (ynr == @dlg_yes)
	[h, bh]
end
	
end	#End Module CurviShear

end	#End Module F6_FredoTools
