=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright © 2011 Fredo6 - Designed and written Jul 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_FredoTools__RemoveLonelyVertices.rb
# Original Date	:   15 Jul 2011 (based on work published in Nov 2009)
# Description	:   Mark all vertices of the selection with a guide point
#                   This is useful to visualize the vertices
# IMPORTANT		:	DO NOT TRANSLATE STRINGS in the source code
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module F6_FredoTools

module RemoveLonelyVertices

#Texts for the module
T7[:MSG_NoLonelyVertices] = "NO Lonely vertex in the selection"
T7[:MSG_Remove] = "Remove"
T7[:MSG_Mark] = "Mark"
T7[:MSG_ApplyRecursively] = "Apply recursively"
T7[:MSG_ActionForLonely] = "Action for %1 lonely vertices?"

#====================================================================================================
#----------------------------------------------------------------------------------------------------
# Plugin Implementation
#----------------------------------------------------------------------------------------------------
#====================================================================================================

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self._execution(symb)
	execute
end

#----------------------------------------------------------------------------------------------------
# Plugin Execution
#----------------------------------------------------------------------------------------------------

def self.init_text
	@color_cube = 'green'

	@menutitle = T7[:PlugName]
	@msg_processing = T6[:MSG_Processing]
	@msg_done = T6[:MSG_Finished]
	@action_lonely = T7[:MSG_ActionForLonely]
	@msg_recursion = T7[:MSG_ApplyRecursively]
	@dlg_remove = T7[:MSG_Remove]
	@dlg_mark = T7[:MSG_Mark]
	@dlg_yes = T6[:T_Yes]
	@dlg_no = T6[:T_No]
	@dlg_yesno = @dlg_yes + '|' + @dlg_no
	@dlg_action = @dlg_remove + '|' + @dlg_mark	
end

#============================================================
# Top Processing method
#============================================================

#Initialize and execute Resolution of Lonely vertices
def self.execute(selection=nil)
	init_text
	@hsh_vertices = {}
	@hsh_entities = {}
	@model = Sketchup.active_model
	selection = @model.selection unless selection
	selection = @model.active_entities if selection.empty?
	
	#Create the structure storing information on lonely vertices
	process_selection selection
			
	#No lonely vertices
	if !@hsh_vertices.values.find { |hsh| hsh.length > 0 }
		UI.messagebox T7[:MSG_NoLonelyVertices]
		return
	end
	
	#Stating Model Operations
	Sketchup.set_status_text @msg_processing, SB_VCB_LABEL
	@model.start_operation @menutitle
	color_edge = @model.rendering_options["EdgeColorMode"]
	@model.rendering_options["EdgeColorMode"] = 0
	unit_cube

	#Showing the lonely vertices with green cubes
	@nb_vertices = 0
	@hsh_vertices.each do |key, hsh|
		entities = @hsh_entities[key]
		hsh.each do |k, info| 
			create_cube_at_vertex(entities, info, @mat)
			@nb_vertices += 1
		end
	end
	
	#Asking for confirmation to proceed
	status = self.dialog
	unless status
		@model.rendering_options["EdgeColorMode"] = color_edge
		@model.abort_operation
		return
	end	
	
	#Removing the lonely vertices
	n = 0
	@hsh_vertices.each do |key, hsh|
		entities = @hsh_entities[key]
		method2 = (key == 0 && hsh.length > 0 && @model.selection.length > 0)
		hsh.each do |k, info|
			n += 1
			Sketchup.set_status_text "#{n}", SB_VCB_VALUE
			
			#Getting info on lonely vertex
			vx, ptx, axes, grp = info
			grp.erase! if grp && grp.valid?
			next unless vx.valid?
			
			#Do nothing if not recursive
			next if @ask_level && key != 0 && @@level == :top
			
			#Marking vertices
			if @@action == :mark
				entities.add_cpoint ptx
				next
			end	
			
			#Removing vertices
			[0, 1].each { |i| e = vx.edges[i] ; e.explode_curve if e.curve }
			if method2
				remove_lonely_vertex_method_2 entities, vx
			else
				remove_lonely_vertex_method_1 entities, info
			end	
		end	
	end
	
	@model.materials.remove @mat
	@model.rendering_options["EdgeColorMode"] = color_edge
	@model.commit_operation
	Sketchup.set_status_text @msg_done, SB_VCB_LABEL
end

#============================================================
# Core algorithms to remove lonely vertices
#============================================================

#Algorithm #1 to remove Lonely vertex 
#  - Draw a small line at vertex and erase the line
#  - Drawback is that it erases all lonely vertices in a sequence of collinear segments
def self.remove_lonely_vertex_method_1(entities, info)
	vx, ptx, axes = info
	return unless axes
	ptend = ptx.offset axes[0], 0.01
	edge = entities.add_line ptx, ptend  
	edge.erase!
end

#Algorithm #2 to remove Lonely vertex 
#  - Create a curve, move its vertices, and create a segment over it
#  - Used for partial selection of colinear segments
def self.remove_lonely_vertex_method_2(entities, vx)
	edges = vx.edges
	v0 = edges[0].other_vertex vx
	v2 = edges[1].other_vertex vx
	id2 = v2.entityID
	lpt = [v0, vx, v2].collect { |v| v.position }
	
	#Creating a curve for the two collinear segments
	grp = entities.add_group
	grp.entities.add_curve lpt
	grp.explode
	curve = vx.edges[0].curve
	
	#Moving the lonely vertex to the extremity
	lpt[1] = lpt[2]
	curve.move_vertices lpt
	e = curve.edges[0]
	e.explode_curve
	
	#Creating a dummy segment in a group and explode it to remove lonely vertex
	grp = entities.add_group
	grp.entities.add_curve [lpt[2], lpt[0]]
	grp.explode
	
	#Updating the valid vertex, vx or v2
	if vx.valid?
		@hsh_vertices[0].each { |k, info| info[0] = vx if info[0] == v2 }
	end	
end

#============================================================
# Managing the dialog box for parameters
#============================================================

@@action = :remove
@@level = :all

def self.dialog
	prompts = []
	results = []
	combo = []
	
	#Flag for conditional asking
	@ask_level = @hsh_vertices.length > 1 && @hsh_vertices[0] && @hsh_vertices[0].length > 0
	
	#Action to be taken
	prompts << @action_lonely.sub("%1", "#{@nb_vertices}")
	results << ((@@action == :remove) ? @dlg_remove : @dlg_mark)
	combo << @dlg_action
	
	#Level of recursion
	if @ask_level
		prompts << @msg_recursion
		results << ((@@level == :all) ? @dlg_yes : @dlg_no)
		combo << @dlg_yesno
	end	
	
	#Invoking the dialog box
	while true
		results = UI.inputbox prompts, results, combo, @menutitle
		return false unless results
		break
	end
	
	@@action = (results[0] == @dlg_remove) ? :remove : :mark
	@@level = (results[1] == @dlg_yes) ? :all : :top if @ask_level
	
	true
end

#============================================================
# Managing Small cubes for marking lonely vertices
#============================================================

#Create once a unit cube
def self.unit_cube
	lfaces = []
	[-1, 1].each { |i| lfaces.push [[-1, -1, i], [1, -1, i], [1, 1, i], [-1, 1, i]] }
	[-1, 1].each { |i| lfaces.push [[i, -1, -1], [i, 1, -1], [i, 1, 1], [i, -1, 1]] }
	[-1, 1].each { |i| lfaces.push [[-1, i, -1], [1, i, -1], [1, i, 1], [-1, i, 1]] }
	@unit_cube = lfaces.collect { |lp| lp.collect { |a| Geom::Point3d.new *a } }
	@mat = @model.materials.add "mat_lonely_vertices"
	@mat.color = @color_cube
	@mat_cur = @model.materials.add "mat_lonely_vertices_cur"
	@mat_cur.color = 'red'
end	
	
#Draw a small red cube at each lonely vertex
def self.create_cube_at_vertex(entities, info, mat)
	vx, ptx, axes = info
	return unless axes
	ptx = vx.position
	size = @model.active_view.pixels_to_model 2, ptx
	tx = Geom::Transformation.axes ptx, *axes
	ts = Geom::Transformation.scaling size
	t = tx * ts
	mesh = Geom::PolygonMesh.new
	@unit_cube.each do |face|
		face3d = face.collect { |pt| t * pt }
		mesh.add_polygon face3d
	end
	grp = entities.add_group
	grp.entities.add_faces_from_mesh mesh, 12, mat, mat
	grp.entities.each { |e| e.material = mat if e.instance_of?(Sketchup::Edge) } 
	info.push grp
end

#============================================================
# Processing the selection recursively
#============================================================

#Explore the selection recursively and store vertex information		
def self.process_selection(lentities, comp=nil)
	hsh, key = get_hash(comp)
	return unless hsh
	lentities.each do |e|
		if e.instance_of?(Sketchup::Edge)
			store_vertices key, lentities, hsh, [e.start, e.end]
		elsif e.instance_of?(Sketchup::Face)
			store_vertices key, lentities, hsh, e.vertices.to_a, true
		elsif e.instance_of?(Sketchup::Group)
			process_selection e.entities, e
		elsif e.instance_of?(Sketchup::ComponentInstance)
			process_selection e.definition.entities, e
		end
	end	
end
	
#Get the hash table to store vertices	
def self.get_hash(comp)
	if comp.instance_of?(Sketchup::ComponentInstance)
		key = comp.definition.entityID
		hsh = @hsh_vertices[key]
		return nil if hsh
		entities = comp.definition.entities
	elsif comp.instance_of?(Sketchup::Group)
		key = comp.entityID
		entities = comp.entities
	else	
		key = 0
		entities = @model.active_entities
	end	
	hsh = @hsh_vertices[key]
	return [hsh, key] if hsh
	
	hsh = @hsh_vertices[key] = {}
	@hsh_entities[key] = entities
	[hsh, key]
end

#Store Lonely vertices vertices	
def self.store_vertices(key, lentities, hsh, lvx, notest=false)
	notest = true if key != 0
	lvx.each do |vx|
		next if hsh[vx.entityID]
		edges = vx.edges
		next if edges.length != 2
		e0 = edges[0]
		e2 = edges[1]
		ptx = vx.position
		v0 = e0.other_vertex(vx)
		v2 = e2.other_vertex(vx)
		pt0 = v0.position
		pt2 = v2.position
		next unless (pt0 == pt2) || (ptx.on_line? [pt0, pt2])
		e0 = vx.edges[0]
		vec = pt0.vector_to pt2
		axes = (vec.valid?) ? vec.axes : nil
		info = [vx, ptx, axes]
		if notest || (lentities.include?(e0) && lentities.include?(e2))
			hsh[vx.entityID] = info
		end
	end
end
	
end	#End Module RemoveLonelyVertices

end	#End Module F6_FredoTools
